/*
* "PS2" Application Framework
*
* University of Abertay Dundee
* May be used for educational purposed only
*
* Author - Dr Henry S Fortuna
*
* $Revision: 1.2 $
* $Date: 2007/08/18 08:07:19 $
*
*/

#include <sps2lib.h>
#include <sps2tags.h>
#include <sps2util.h>
#include <sps2regstructs.h>
#include <assert.h>
#include <memory.h>
#include <signal.h>
#include "PS2Defines.h"
#include "ps2maths.h"
#include "ps2matrix4x4.h"
#include "pad.h"
#include "sps2wrap.h"
#include "dma.h"

// These two external function pointers are exported by the VU code
// and we can now use them as pointers to the compiled VU data
extern "C" void VU_vu1code_start(void);
extern "C" void VU_vu1code_end(void);

bool g_bLoop = true;

void sig_handle(int sig)
{
	g_bLoop = false;
}

int main(void)
{
	// Initialise control pad 0
	if(!pad_init(PAD_0, PAD_INIT_LOCK | PAD_INIT_ANALOGUE | PAD_INIT_PRESSURE))
	{
		printf("Failed to initialise control pad\n");
		pad_cleanup(PAD_0);
		exit(0);
	}	

	// Initialise the memory
	SPS2Manager.Initialise(4096);	// 4096 * 4K Pages = 16MB Total
	VIFStaticDMA.Initialise(1536);	// 1536 * 4K Pages = 6MB Static DMA
	VIFDynamicDMA.Initialise(256);	// 256 * 4K Pages * 2 Buffers =
									// 2MB Dynamic DMA

	// Find out the length of the code in double words
	// We use double words since code is made up of many
	// pairs of two (1 upper, and 1 lower) VU instructions. 
	// Each instruction is 32 bits, so VU code must be a
	// multiple of 64 bits.
	uint32 iCodeLengthInDWORDs = (((uint32)&VU_vu1code_end - (uint32)&VU_vu1code_start) + 7) >> 3;
	// Copy the code to VU1_MICRO_MEM
	memcpy(VU1_MICRO_MEM, &VU_vu1code_start, iCodeLengthInDWORDs * 8);
	
	
	// Register our signal function for every possible signal (i.e. ctrl + c)
	for(int sig=0; sig<128; sig++)
		signal(sig, sig_handle);


	// Build a projection matrix
	float n = 2.0f;		// Near plane
	float f = 2000.0f;	// Far plane
	float FovYdiv2 = DegToRad(60.0f) * 0.5f; // 60 degree FOV
	float fAspect = 4.0f / 3.0f;			 // Aspect ratio
	float cotFOV = 1.0f / (Sin(FovYdiv2) / Cos(FovYdiv2));
	
	// We will be projecting to the 4096 wide drawing area
	// Look also at the VU projection code
	float w = cotFOV * (sps2UScreenGetWidth() / 4096.0f) / fAspect;
	float h = cotFOV * (sps2UScreenGetHeight() / 4096.0f);

	Matrix4x4 matProj
		(w, 0, 0, 0,
		 0, -h, 0, 0,
		 0, 0, (f+n) / (f-n), -1,
		 0, 0, (2*f*n) / (f-n), 0);
	
	//matProj.DumpMatrix4x4("matProj");

	Matrix4x4 matView;
	matView.LookAt(Vector4(0, 0, 0, 1), Vector4(0, 0, -10.0f, 1), Vector4(0, 1, 0, 1));
	//matView.DumpMatrix4x4("matView");

	Matrix4x4 matViewProj = matView * matProj;
	//matViewProj.DumpMatrix4x4("matViewProj");
	
	// Get the address of static data that we can call to later.
	int iStaticAddr = VIFStaticDMA.GetPointer();

	// Drawing 2 triangles.
	int iTriangles = 2;
	// There are three vertices per triangle
	int iVerts = iTriangles * 3;

	// We want to unpack 2 quad words of misc data, and then 2QW per vertex starting at
	// VU mem location 4 (we will upload the World matrix to the first 4QW).
	// Unpack at location 4
	// Each vertex is 2 qwords in size
	VIFStaticDMA.AddUnpack(V4_32, 4, 2 + iVerts * 2);
	VIFStaticDMA.AddVector(Vector4(2048.0f, 2048.0f, ((float)0xFFFFFF) / 32.0f, 0.0f));


	// Add the GIFTag
	VIFStaticDMA.Add128(GS_GIFTAG_BATCH(	iVerts,						// NLOOP
	 										1,							// EOP
	 										1,							// PRE
	 										GS_PRIM(PRIM_TRI,		 	// PRIM
													PRIM_IIP_GOURAUD, 
													PRIM_TME_OFF,
													PRIM_FGE_OFF, 
													PRIM_ABE_OFF, 
													PRIM_AA1_OFF, 
													PRIM_FST_UV, 
													PRIM_CTXT_CONTEXT1, 
													PRIM_FIX_NOFIXDDA),					
	 										GIF_FLG_PACKED,				// FLG
	 										GS_BATCH_2(	GIF_REG_RGBAQ, 	// BATCH
	 													GIF_REG_XYZ2)));


	// Add the quad
	VIFStaticDMA.Add128(PACKED_RGBA(0x80, 0x00, 0x00, 0x80));	// Colour
	VIFStaticDMA.AddVector(Vector4(-3.0f, 3.0f, 0.0f, 1.0f));	// Vert
	VIFStaticDMA.Add128(PACKED_RGBA(0x80, 0x00, 0x00, 0x80));	// Colour
	VIFStaticDMA.AddVector(Vector4(3.0f,  3.0f, 0.0f, 1.0f));	// Vert
	VIFStaticDMA.Add128(PACKED_RGBA(0x80, 0x00, 0x00, 0x80));	// Colour
	VIFStaticDMA.AddVector(Vector4(-3.0f,-3.0f, 0.0f, 1.0f));	// Vert
	VIFStaticDMA.Add128(PACKED_RGBA(0x00, 0x00, 0x80, 0x80));	// Colour
	VIFStaticDMA.AddVector(Vector4(-3.0f,-3.0f, 0.0f, 1.0f));	// Vert
	VIFStaticDMA.Add128(PACKED_RGBA(0x00, 0x00, 0x80, 0x80));	// Colour
	VIFStaticDMA.AddVector(Vector4(3.0f,  3.0f, 0.0f, 1.0f));	// Vert
	VIFStaticDMA.Add128(PACKED_RGBA(0x00, 0x00, 0x80, 0x80));	// Colour
	VIFStaticDMA.AddVector(Vector4(3.0f, -3.0f, 0.0f, 1.0f));	// Vert
	
	// Flush to make sure VU1 isn't doing anything.
	VIFStaticDMA.Add32(FLUSH);
	// Then run the microcode.
	VIFStaticDMA.Add32(MSCALL(0));

	// Return back to the dynamic chain (note that we aren't aligned on a QWORD boundary, but
	// it doesn't matter because the packet class will sort that out for us).
	VIFStaticDMA.DMARet();

	// Set up the DMA packet to clear the screen. We want to clear to blue.
	SPS2Manager.InitScreenClear(0x00, 0x05, 0x10);
	
	while(g_bLoop)
	{
		VIFDynamicDMA.Fire();
		
		pad_update(PAD_0);
	
		// Set up some axes rotations
		static float RotX = 0.0f;
		static float RotY = 0.0f;
		static float RotZ = 0.0f;
		
		// change rotations for the next frame
		RotX += 0.021f;
		RotY += 0.03f;
		RotZ += 0.011f;
		
		// The required matrices
		Matrix4x4 matWorld, matTrans, matRotX, matRotY, matRotZ;

		// Set up a translation matrix
		matTrans.Translation(0.0f, 0.0f, -15.0f);
				
		// Get the rotation matrices
		matRotX.RotationX(RotX);
		matRotY.RotationY(RotY);
		matRotZ.RotationZ(RotZ);

		matWorld = matRotZ * matRotY * matRotX * matTrans;

		// You should be familiar with the idea of building 3D transforms from
		// the World * View * Projection matrices
		Matrix4x4 matWVP = matWorld * matViewProj;
		
		SPS2Manager.BeginScene();
		
		// We have to upload the matrix in the dynamic buffer
		// because it change every frame.
		VIFDynamicDMA.AddUnpack(V4_32, 0, 4); // Unpack 4QW to VU mem addr 0
		VIFDynamicDMA.AddMatrix(matWVP);
		
		// We can simply call the rest, which means the CPU doesn't have to do any work.
		VIFDynamicDMA.DMACall(iStaticAddr);
				
		SPS2Manager.EndScene();
		
		// Check for exit condition
		if((pad[0].buttons & PAD_START)&&(pad[0].buttons & PAD_SELECT)) g_bLoop = false;
	}

	pad_cleanup(PAD_0);
	
	return 0;
	
}

